@@ -2,9 +2,15 @@
 A collection of reference ISP algorithms.
 """
 
-import time             # built-in library
-import numpy as np      # pip install numpy
-import cv2              # pip install opencv-python
+import time                  # built-in library
+import warnings              # built-in library
+from typing import Tuple     # built-in library
+from typing import Sized     # built-in library
+from typing import Optional  # built-in library
+from typing import Callable  # built-in library
+import numpy as np           # pip install numpy
+import colorio               # pip install colorio
+import cv2                   # pip install opencv-python
 
 
 class Algorithms:
@@ -15,23 +21,21 @@ class Algorithms:
     def __init__(self, verbose=False):
         """
         Initialize self. If verbose is True, progress information will be printed
-        to stdout, unless the per-function verbose flag is set to False.
+        to stdout.
         """
         self.verbose = verbose
-        self._verbose = verbose
 
-    def clip(self, frame, lo=0.0, hi=1.0, verbose=True):
+    def clip(self, frame: np.ndarray, lo: float = 0.0, hi: float = 1.0) -> np.ndarray:
         """
         Clip all pixels in the given frame to [lo, hi]. The frame may be in either
         RGB or raw format.
         """
         t0 = time.time()
         frame_out = np.clip(frame, lo, hi)
-        self._verbose = self.verbose and verbose
-        self._vprint(f"{_elapsed(t0)} - clipping from {_minmax(frame)} to [{lo:.2f}, {hi:.2f}]")
+        self._vprint(f"{_elapsed(t0)} - clipping from {self._minmax(frame)} to [{lo:.2f}, {hi:.2f}]")
         return frame_out
 
-    def bayer_split(self, frame):
+    def bayer_split(self, frame: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
         """
         Split the given Bayer frame into four single-color frames.
         """
@@ -42,7 +46,11 @@ class Algorithms:
         ch4 = frame[1::2, 1::2]
         return ch1, ch2, ch3, ch4
 
-    def bayer_combine(self, ch1, ch2, ch3, ch4):
+    def bayer_combine(self,
+                      ch1: np.ndarray,
+                      ch2: np.ndarray,
+                      ch3: np.ndarray,
+                      ch4: np.ndarray) -> np.ndarray:
         """
         Interleave the given Bayer channels into a complete frame.
         """
@@ -54,12 +62,12 @@ class Algorithms:
         frame[1::2, 1::2] = ch4
         return frame
 
-    def downsample(self, frame, iterations=1, verbose=True):
+    def downsample(self, frame: np.ndarray, iterations: int = 1) -> np.ndarray:
         """
         Downsample the given RGB/Bayer frame by a factor of two in both directions,
         that is, to a quarter of its original size. This is repeated a given number
         of iterations. If iterations is zero, the frame is returned untouched. See
-        resample() for arbitrary resizing with proper interpolation in RGB domain.
+        resize() for arbitrary resizing with proper interpolation in RGB domain.
         """
         if iterations >= 1:
             t0 = time.time()
@@ -78,11 +86,10 @@ class Algorithms:
                     frame = self.bayer_combine(ch1, ch2, ch3, ch4)
             imgh, imgw = frame.shape[:2]
             factor = f"{2**iterations} x {2**iterations}"
-            self._verbose = self.verbose and verbose
             self._vprint(f"{_elapsed(t0)} - {mode} downsampling [{factor}] from {orgw} x {orgh} to {imgw} x {imgh}")
         return frame
 
-    def resize(self, frame, target_width, target_height, verbose=True):
+    def resize(self, frame: np.ndarray, target_width: int, target_height: int) -> np.ndarray:
         """
         Resize the given RGB frame to the given target width and height using Lanczos
         interpolation. If target width and height are the same as current width and
@@ -94,11 +101,14 @@ class Algorithms:
         if (dstw, dsth) != (orgw, orgh):
             frame = cv2.resize(frame, (dstw, dsth), cv2.INTER_LANCZOS4)
             dsth, dstw = frame.shape[:2]
-            self._verbose = self.verbose and verbose
             self._vprint(f"{_elapsed(t0)} - resizing [Lanczos] from {orgw} x {orgh} to {dstw} x {dsth}")
         return frame
 
-    def linearize(self, frame, blacklevel=None, whitelevel=None, num_clipped=1000, verbose=True):
+    def linearize(self,
+                  frame: np.ndarray,
+                  blacklevel: float = None,
+                  whitelevel: float = None,
+                  num_clipped: int = 1000) -> np.ndarray:
         """
         Linearize the given frame such that pixels are first clipped to the range
         [BL, WL] and then remapped to [0, 1], where BL and WL are the given black
@@ -112,44 +122,62 @@ class Algorithms:
             t0 = time.time()
             percentile = num_clipped / frame.size * 100.0
             blacklevel = np.percentile(frame, percentile)
-            self._verbose = self.verbose and verbose
             self._vprint(f"{_elapsed(t0)} - estimating black level: {percentile:5.2f}th percentile = {blacklevel:.2f}")
         if whitelevel is None:
             t0 = time.time()
             percentile = (1.0 - num_clipped / frame.size) * 100.0
             whitelevel = np.percentile(frame, percentile)
-            self._verbose = self.verbose and verbose
             self._vprint(f"{_elapsed(t0)} - estimating white level: {percentile:5.2f}th percentile = {whitelevel:.2f}")
         if (blacklevel, whitelevel) != (0, 1):
-            frame = self.clip(frame, blacklevel, whitelevel, verbose)
+            frame = self.clip(frame, blacklevel, whitelevel)
             t0 = time.time()
             frame -= blacklevel
             frame /= whitelevel - blacklevel
-            self._verbose = self.verbose and verbose
             self._vprint(f"{_elapsed(t0)} - linearizing from [{blacklevel:.2f}, {whitelevel:.2f}] to [0, 1]")
         return frame
 
-    def demosaic(self, frame, bayer_pattern, verbose=True):
+    def demosaic(self,
+                 frame: np.ndarray,
+                 bayer_pattern: str,
+                 downsample: bool = False) -> np.ndarray:
         """
         Demosaic the given sensor raw frame using the Edge Aware Demosaicing (EAD)
         algorithm. Bayer order must be specified by the caller, and must be "RGGB",
         "GBRG", "BGGR", or "GRBG". The frame must be in floating-point format with
         all pixel values in the [0, 1] range.
+
+        If the 'downsample' flag is True, RGB values are picked from the raw Bayer
+        pattern as-is, without any interpolation other than averaging the greens.
+        This reduces the size of the image by a factor of 2 x 2. In 'downsample'
+        mode, pixel values need not be in [0, 1] range.
         """
         t0 = time.time()
-        assert np.all((frame >= 0.0) * (frame <= 1.0)), "demosaic() requires pixel values in range [0, 1]"
-        bayer_to_cv2 = {"RGGB": cv2.COLOR_BAYER_BG2RGB_EA,
-                        "GBRG": cv2.COLOR_BAYER_GR2RGB_EA,
-                        "BGGR": cv2.COLOR_BAYER_RG2RGB_EA,
-                        "GRBG": cv2.COLOR_BAYER_GB2RGB_EA}
-        frame = (frame * 65535).astype(np.uint16)
-        frame = cv2.cvtColor(frame, bayer_to_cv2[bayer_pattern.upper()])
-        frame = frame / 65535.0
-        self._verbose = self.verbose and verbose
-        self._vprint(f"{_elapsed(t0)} - demosaicing [EAD, {bayer_pattern}]: range = {_minmax(frame)}")
+        if not downsample:
+            assert np.all((frame >= 0.0) * (frame <= 1.0)), "demosaic() requires pixel values in range [0, 1]"
+            bayer_to_cv2 = {"RGGB": cv2.COLOR_BAYER_BG2RGB_EA,
+                            "GBRG": cv2.COLOR_BAYER_GR2RGB_EA,
+                            "BGGR": cv2.COLOR_BAYER_RG2RGB_EA,
+                            "GRBG": cv2.COLOR_BAYER_GB2RGB_EA}
+            frame = (frame * 65535).astype(np.uint16)
+            frame = cv2.cvtColor(frame, bayer_to_cv2[bayer_pattern.upper()])
+            frame = frame / 65535.0
+            method = "EAD"
+        else:
+            channels = self.bayer_split(frame)
+            bayer_to_index = {"RGGB": [0, 1, 2, 3],
+                              "GBRG": [2, 0, 3, 1],
+                              "BGGR": [3, 1, 2, 0],
+                              "GRBG": [1, 0, 3, 2]}
+            indices = bayer_to_index[bayer_pattern.upper()]
+            r = channels[indices[0]]
+            g = (channels[indices[1]] + channels[indices[2]]) / 2.0
+            b = channels[indices[3]]
+            frame = np.dstack((r, g, b))
+            method = "downsample"
+        self._vprint(f"{_elapsed(t0)} - demosaicing [{method}, {bayer_pattern}]: range = {self._minmax(frame)}")
         return frame
 
-    def lsc(self, frame, lscmap, verbose=True):
+    def lsc(self, frame: np.ndarray, lscmap: np.ndarray) -> np.ndarray:
         """
         Multiply the given RGB/raw frame by the given lens shading correction (LSC)
         map. If the frame is in Bayer raw format, the LSC map must have the same
@@ -180,11 +208,10 @@ class Algorithms:
                     gains = np.amax(lscmap, axis=(0, 1))
                 if lscmap.ndim == 2:  # assume Bayer raw, ignore grayscale
                     gains = np.array([np.amax(c) for c in self.bayer_split(lscmap)])
-                self._verbose = self.verbose and verbose
-                self._vprint(f"{_elapsed(t0)} - applying LSC with max gains {gains}: range = {_minmax(frame)}")
+                self._vprint(f"{_elapsed(t0)} - applying LSC with max gains {gains}: range = {self._minmax(frame)}")
         return frame
 
-    def wb(self, frame, gains, verbose=True):
+    def wb(self, frame: np.ndarray, gains: Sized) -> np.ndarray:
         """
         Multiply the RGB channels of the given frame by the given white balance
         coefficients. If there are only two coefficients instead of three, they
@@ -197,11 +224,10 @@ class Algorithms:
             gains = np.insert(gains, 1, 1.0) if len(gains) == 2 else gains
             frame = frame * gains
             with np.printoptions(formatter={'float': lambda x: f"{x:.3f}"}):
-                self._verbose = self.verbose and verbose
-                self._vprint(f"{_elapsed(t0)} - applying WB gains {gains}: range = {_minmax(frame)}")
+                self._vprint(f"{_elapsed(t0)} - applying WB gains {gains}: range = {self._minmax(frame)}")
         return frame
 
-    def ccm(self, frame, matrix, verbose=True):
+    def ccm(self, frame: np.ndarray, matrix: np.ndarray) -> np.ndarray:
         """
         Apply the given color correction matrix on the given RGB frame. Input
         colors are clipped to [0, 1] to avoid "pink sky" artifacts caused by the
@@ -210,16 +236,15 @@ class Algorithms:
         recovery. If matrix is None, the frame is returned untouched.
         """
         if matrix is not None:
-            frame = self.clip(frame, 0, 1, verbose=False)
+            frame = self.clip(frame, 0, 1)
             t0 = time.time()
-            frame = np.matmul(frame, matrix.T)
+            frame = frame @ matrix.T
             with np.printoptions(formatter={'float': lambda x: f"{x:.2f}"}):
                 sums = f"with column sums {np.sum(matrix, axis=0).T}"
-                self._verbose = self.verbose and verbose
-                self._vprint(f"{_elapsed(t0)} - applying CCM {sums}: range = {_minmax(frame)}")
+                self._vprint(f"{_elapsed(t0)} - applying CCM {sums}: range = {self._minmax(frame)}")
         return frame
 
-    def tonemap(self, frame, mode="Reinhard", verbose=True):
+    def tonemap(self, frame: np.ndarray, mode: str = "Reinhard") -> np.ndarray:
         """
         Apply Reinhard tonemapping on the given RGB frame, compressing the range
         [0, N] to [0, 1]. Negative values are clipped to zero. This algorithm is
@@ -227,16 +252,18 @@ class Algorithms:
         """
         if mode is not None:
             if mode == "Reinhard":
-                frame = self.clip(frame, 0, np.inf, verbose)
+                frame = self.clip(frame, 0, np.inf)
                 t0 = time.time()
                 frame = frame.astype(np.float32)  # can't handle float64
                 algo = cv2.createTonemapReinhard(gamma=1.0, intensity=0.0, light_adapt=0.0, color_adapt=0.0)
                 frame = algo.process(frame)
-                self._verbose = self.verbose and verbose
-                self._vprint(f"{_elapsed(t0)} - tonemapping [{mode}]: range = {_minmax(frame)}")
+                self._vprint(f"{_elapsed(t0)} - tonemapping [{mode}]: range = {self._minmax(frame)}")
         return frame
 
-    def chroma_denoise(self, frame, strength=6, winsize=17, verbose=True):
+    def chroma_denoise(self,
+                       frame: np.ndarray,
+                       strength: int = 6,
+                       winsize: int = 17) -> np.ndarray:
         """
         Apply non-local means denoising (Buades et al. 2011) on the given RGB frame.
         Input colors are clipped to [0, 1] prior to denoising. Increasing the values
@@ -245,77 +272,120 @@ class Algorithms:
         """
         if strength > 0:
             maxval, dtype = (255, np.uint8)  # OpenCV denoising can't handle 16-bit color
-            frame = self.clip(frame * maxval + 0.5, 0, maxval, verbose)
+            frame = self.clip(frame * maxval + 0.5, 0, maxval)
             t0 = time.time()
             frame = frame.astype(dtype)
             frame = cv2.fastNlMeansDenoisingColored(frame, h=0, hColor=strength, searchWindowSize=winsize)
             frame = frame.astype(np.float32) / maxval
-            self._verbose = self.verbose and verbose
-            self._vprint(f"{_elapsed(t0)} - chroma denoising [s={strength:.2f}, w={winsize}]: range = {_minmax(frame)}")
+            self._vprint(f"{_elapsed(t0)} - chroma denoising [s={strength:.2f}, w={winsize}]: range = {self._minmax(frame)}")
         return frame
 
-    def gamma(self, frame, mode="sRGB", lut=None, verbose=True):
+    def saturate(self,
+                 frame: np.ndarray,
+                 booster: Optional[Callable[[np.ndarray], np.ndarray]]) -> np.ndarray:
+        """
+        Apply the caller-provided boost function on the given RGB frame. The input
+        frame is converted to HSL color space and the S channel given as the sole
+        input to the boost function. Input RGB colors are clipped to [0, 1] before
+        converting to HSL. If booster is None, the frame is returned untouched.
+
+        Example:
+          img = pipe.saturate(img, lambda x: x ** 0.75)
+        """
+        if booster is not None:
+            t0 = time.time()
+            hsl = _transform_srgb_to_hsl(frame)  # HSL is in planar form
+            hsl[1] = booster(hsl[1])  # saturation boost
+            hsl[1] = np.clip(hsl[1], 0, 1)  # may be -eps or 1+eps
+            frame = _transform_hsl_to_srgb(hsl)  # RGB back in interleaved form
+            self._vprint(f"{_elapsed(t0)} - applying HSL saturation boost: range = {self._minmax(frame)}")
+        return frame
+
+    def gamma(self,
+              frame: np.ndarray,
+              mode: str = "sRGB",
+              lut: np.ndarray = None) -> np.ndarray:
         """
         Apply rec709 or sRGB gamma or a custom tone curve on the given frame.
-        In "LUT" mode, a look-up table with N entries must be provided, where
-        N equals 2 ** bpp (e.g., N=4096 for 12-bit colors). This algorithm is
-        format-agnostic. If mode is None, the frame is returned untouched.
+        Input colors are clipped to [0, 1] to avoid any arithmetic exceptions.
+        In "LUT" mode, the frame is quantized to match the number of entries N
+        in the look-up table; for example, if N=64, the output frame will have
+        6-bit colors (and severe banding). This algorithm is format-agnostic.
+        If mode is None, the frame is returned untouched.
         """
         if mode is not None:
             realmode = mode
-            frame = self.clip(frame, 0, 1, verbose)  # can't handle values outside of [0, 1]
+            frame = self.clip(frame, 0, 1)  # can't handle values outside of [0, 1]
             t0 = time.time()
             if realmode in ["sRGB", "rec709"]:
                 bpp = 14
-                N = 2 ** bpp
                 maxval = 2 ** bpp - 1
-                lut = np.linspace(0, 1, N)
+                lut = np.linspace(0, 1, 2 ** bpp)
                 mode = "LUT"
-            if realmode == "sRGB":
-                srgb_lo = 12.92 * lut
-                srgb_hi = 1.055 * np.power(lut, 1.0/2.4) - 0.055
-                threshold_mask = (lut > 0.0031308)
-                lut = srgb_hi * threshold_mask + srgb_lo * (~threshold_mask)
-                lut = lut * maxval
-            if realmode == "rec709":
-                srgb_lo = 4.5 * lut
-                srgb_hi = 1.099 * np.power(lut, 0.45) - 0.099
-                threshold_mask = (lut > 0.018)
-                lut = srgb_hi * threshold_mask + srgb_lo * (~threshold_mask)
-                lut = lut * maxval
+                if realmode == "sRGB":
+                    srgb_lo = 12.92 * lut
+                    srgb_hi = 1.055 * np.power(lut, 1.0/2.4) - 0.055
+                    threshold_mask = (lut > 0.0031308)
+                    lut = srgb_hi * threshold_mask + srgb_lo * (~threshold_mask)
+                    lut = lut * maxval
+                if realmode == "rec709":
+                    srgb_lo = 4.5 * lut
+                    srgb_hi = 1.099 * np.power(lut, 0.45) - 0.099
+                    threshold_mask = (lut > 0.018)
+                    lut = srgb_hi * threshold_mask + srgb_lo * (~threshold_mask)
+                    lut = lut * maxval
             if mode == "LUT":
+                assert lut is not None
                 maxval = len(lut) - 1
-                frame = self.quantize(frame, maxval, verbose=False)  # [0, 1] ==> [0, maxval]
-                frame = lut[frame]                                   # [0, maxval] ==> [0, maxval]
-                frame = frame / maxval                               # [0, maxval] ==> [0, 1]
-            self._verbose = self.verbose and verbose
-            self._vprint(f"{_elapsed(t0)} - applying gamma curve [{realmode}]: range = {_minmax(frame)}")
+                frame = self.quantize(frame, maxval)  # [0, 1] ==> [0, maxval]
+                frame = lut[frame]                    # [0, maxval] ==> [0, maxval]
+                frame = frame / maxval                # [0, maxval] ==> [0, 1]
+            self._vprint(f"{_elapsed(t0)} - applying gamma curve [{realmode}]: range = {self._minmax(frame)}")
         return frame
 
-    def quantize(self, frame, maxval, dtype=np.uint16, verbose=True):
+    def quantize(self,
+                 frame: np.ndarray,
+                 maxval: int,
+                 dtype: np.dtype = np.uint16) -> np.ndarray:
         """
         Clip the given frame to [0, 1], rescale it to [0, maxval], and convert
         it to the given dtype with proper rounding. This algorithm is format-
         agnostic.
         """
         t0 = time.time()
-        frame = self.clip(frame * maxval + 0.5, 0, maxval, verbose=False)
+        frame = self.clip(frame * maxval + 0.5, 0, maxval)
         frame = frame.astype(dtype)
-        self._verbose = self.verbose and verbose
         self._vprint(f"{_elapsed(t0)} - quantizing from float [0, 1] to {np.dtype(dtype).name} [0, {maxval}]")
         return frame
 
-    def _vprint(self, message, **kwargs):
-        if self._verbose:
+    def _vprint(self, message: str, **kwargs):
+        if self.verbose:
             print(message, **kwargs)
 
+    def _minmax(self, frame: np.ndarray) -> str:
+        if self.verbose:
+            minmax_str = f"[{np.min(frame):.2f}, {np.max(frame):.2f}]"
+            return minmax_str
 
-def _elapsed(t0):
+
+def _transform_srgb_to_hsl(frame: np.ndarray) -> np.ndarray:
+    rgb = frame.swapaxes(-1, 0)  # (H, W, 3) ==> (3, W, H)
+    rgb = np.clip(rgb, 0.001, 0.999)  # colorio.HSL() can't handle 0.0 and 1.0
+    with warnings.catch_warnings():
+        warnings.simplefilter("ignore", category=RuntimeWarning)
+        hsl = colorio.HSL().from_srgb1(rgb)
+        hsl = np.nan_to_num(hsl)
+    return hsl
+
+
+def _transform_hsl_to_srgb(frame: np.ndarray) -> np.ndarray:
+    rgb = colorio.HSL().to_srgb1(frame)
+    rgb = rgb.swapaxes(-1, 0)  # (3, W, H) => (H, W, 3)
+    rgb = np.ascontiguousarray(rgb)
+    return rgb
+
+
+def _elapsed(t0: float) -> str:
     elapsed = (time.time() - t0) * 1000
-    elapsed = f"{elapsed:8.2f} ms"
-    return elapsed
-
-
-def _minmax(frame):
-    minmax = f"[{np.min(frame):.2f}, {np.max(frame):.2f}]"
-    return minmax
+    elapsed_str = f"{elapsed:8.2f} ms"
+    return elapsed_str
