@@ -31,8 +31,9 @@ class RegressionTest(unittest.TestCase):
         raw = alg.demosaic(raw, bayer_pattern)
         raw = alg.wb(raw, wb)
         raw = alg.ccm(raw, ccm)
-        raw = alg.gamma(raw, gamma_mode)
-        raw = alg.quantize(raw, maxval, expected.dtype)
+        raw = alg.saturate(raw, lambda x: x ** 0.5)
+        raw = rawpipe.verbose.gamma(raw, gamma_mode)
+        raw = rawpipe.silent.quantize(raw, maxval, expected.dtype)
         self.assertEqual(np.count_nonzero(expected - raw), 0)
 
 
